﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using Windows.Devices.Enumeration;
using Windows.Devices.Midi;
using System.Threading.Tasks;
using Windows.UI;
using Windows.ApplicationModel;
using Windows.UI.Xaml.Shapes;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=402352&clcid=0x409

namespace MidiKeyboard
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public sealed partial class MainPage : Page
    {
        // MidiOutPort of the output MIDI device
        IMidiOutPort midiOutPort;

        // MIDI message to change the instument
        IMidiMessage instrumentChange;

        // Integer to define the frequecy interval of an instrument
        int octaveInterval = 0;

        // The MIDI channel used for the MIDI output device
        byte channel = 0;

        // The MIDI velocity used for the MIDI output device
        byte velocity = 127;

        public MainPage()
        {
            this.InitializeComponent();

            // Set the initial instrument of the keyboard to Acoustic Grand Piano in channel 0
            instrumentChange = new MidiProgramChangeMessage(0, 0);

            // Get the MIDI output device
            GetMidiOutputDevicesAsync();

            // Reopen the midi connection when resuming after a suspension
            Application.Current.Resuming += new EventHandler<Object>(App_Resuming);

            // Close the midi connection on suspending the app
            Application.Current.Suspending += new SuspendingEventHandler(App_Suspending);
        }


        /// <summary>
        /// Eventhandler to clean up the MIDI connection, when the app is suspended.
        /// The object midiOutPort is disposed and set to null
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void App_Suspending(object sender, SuspendingEventArgs e)
        {
            try
            {
                // Dispose the Midi Output port
                midiOutPort.Dispose();

                // Set the midiOutPort to null
                midiOutPort = null;
            }
            catch
            {
                // Do noting. A cleanup has already been made
            }
        }

        /// <summary>
        /// Eventhandler to restore connection to the MIDI device when app is resuming after suspension.
        /// The method EnumerateMidiOutputDevices() will be called.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private async void App_Resuming(object sender, object e)
        {
            // Call method to set up output midi device
            await EnumerateMidiOutputDevices();
        }

        /// <summary>
        /// Method to call EnumarateMidiOutputDevice from the contructor of the MainPage.
        /// EnumerateMidiOutputDevices() has to be called via a await call, and that cannot be done from the constructor.
        /// </summary>
        private async void GetMidiOutputDevicesAsync()
        {
            await EnumerateMidiOutputDevices();
        }


        /// <summary>
        /// Method to enumrate all MIDI output devices connected and to setup the the first MIDI output device found.
        /// </summary>
        /// <returns></returns>
        private async Task EnumerateMidiOutputDevices()
        {
            // Create the query string for finding all MIDI output devices using MidiOutPort.GetDeviceSelector()
            string midiOutportQueryString = MidiOutPort.GetDeviceSelector();

            // Find all MIDI output devices and collect it in a DeviceInformationCollection using FindAllAsync
            DeviceInformationCollection midiOutputDevices = await DeviceInformation.FindAllAsync(midiOutportQueryString);

            // If the size of the midiOutputDevice colloction is xero,
            // set the StatusTextBlock foreground color to red and the text property to "No MIDI output devices found"
            // and return.
            // Else set the StatusTextBlock foreground color to green and the text property to midiOutputdevices[0].Name
            if (midiOutputDevices.Count == 0)
            {
                // Set the StatusTextBlock foreground color to red
                StatusTextBlock.Foreground = new SolidColorBrush(Colors.Red);

                // Set the StatusTextBlock text to "No MIDI output devices found"
                StatusTextBlock.Text = "No MIDI output devices found";
                return;
            }
            else
            {
                // Set the StatusTextBlock foreground color to green
                StatusTextBlock.Foreground = new SolidColorBrush(Colors.Green);

                // Set the StatusTextBlock text to the name of the first item in midiOutputDevices collection
                StatusTextBlock.Text = midiOutputDevices[0].Name;
            }

            // Create an instance of DeviceInformation and set it to the first midi device in DeviceInformationCollection, midiOutputDevices
            DeviceInformation devInfo = midiOutputDevices[0];

            // Return if DeviceInformation, devInfo, is null
            if (devInfo == null)
            {
                // Set the midi status TextBlock
                StatusTextBlock.Foreground = new SolidColorBrush(Colors.Red);
                StatusTextBlock.Text = "No device information of MIDI output";
                return;
            }

            // Set the IMidiOutPort for the output midi device by calling MidiOutPort.FromIdAsync passing the Id property of the DevicInformation
            midiOutPort = await MidiOutPort.FromIdAsync(devInfo.Id);

            // Return if midiOutPort is null
            if (midiOutPort == null)
            {
                // Set the midi status TextBlock
                StatusTextBlock.Foreground = new SolidColorBrush(Colors.Red);
                StatusTextBlock.Text = "Unable to create MidiOutPort from output device";
                return;
            }

            // Send the Program Change midi message to port of the output midi device
            // to set the initial instrument to Acoustic Grand Piano.
            midiOutPort.SendMessage(instrumentChange);
        }

        /// <summary>
        /// Eventhandler for key released at the keyboard.
        /// The specific key is extracted and the according pitch is found.
        /// A MIDI message Note Off is created from the channel field and velocity field.
        /// The Note Off message is send to the MIDI output device
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Key_PointerReleased(object sender, PointerRoutedEventArgs e)
        {
            // Field to hold the pitch of the note
            byte note;

            // Extract the key being pressed
            Rectangle keyPressed = (Rectangle)sender;

            // Get the name of the key and store it in a string, keyPressedName
            string keyPressedName = keyPressed.Name;

            // Switch/Case to set the pitch depending of the key pressed
            switch (keyPressedName)
            {
                case "CKey":
                    note = (byte)(60 + (octaveInterval * 12));
                    break;
                case "CSharpKey":
                    note = (byte)(61 + (octaveInterval * 12));
                    break;
                case "DKey":
                    note = (byte)(62 + (octaveInterval * 12));
                    break;
                case "DSharpKey":
                    note = (byte)(63 + (octaveInterval * 12));
                    break;
                case "EKey":
                    note = (byte)(64 + (octaveInterval * 12));
                    break;
                default:
                    note = (byte)(60 + (octaveInterval * 12));
                    break;
            }

            // Create the Note Off message to send to the MIDI output device
            IMidiMessage midiMessageToSend = new MidiNoteOffMessage(channel, note, velocity);

            // Send the Note Off MIDI message to the midiOutPort
            midiOutPort.SendMessage(midiMessageToSend);
        }

        /// <summary>
        /// Eventhandler for key pressed at the keyboard.
        /// The specific key is extracted and the according pitch is found.
        /// A MIDI message Note On is created from the channel field and velocity field.
        /// The Note On message is send to the MIDI output device
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Key_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            // Field to hold the pitch of the note
            byte note;
            
            // Extract the key being pressed
            Rectangle keyPressed = (Rectangle)sender;

            // Get the name of the key and store it in a string, keyPressedName
            string keyPressedName = keyPressed.Name;

            // Switch/Case to set the pitch depending of the key pressed
            switch (keyPressedName)
            {
                case "CKey":
                    note = (byte)(60 + (octaveInterval * 12));
                    break;
                case "CSharpKey":
                    note = (byte)(61 + (octaveInterval * 12));
                    break;
                case "DKey":
                    note = (byte)(62 + (octaveInterval * 12));
                    break;
                case "DSharpKey":
                    note = (byte)(63 + (octaveInterval * 12));
                    break;
                case "EKey":
                    note = (byte)(64 + (octaveInterval * 12));
                    break;
                default:
                    note = (byte)(60 + (octaveInterval * 12));
                    break;
            }

            // Create the Note On message to send to the MIDI output device
            IMidiMessage midiMessageToSend = new MidiNoteOnMessage(channel, note, velocity);

            // Send the Note On MIDI message to the midiOutPort
            midiOutPort.SendMessage(midiMessageToSend);     
        }

        /// <summary>
        /// Eventhandler for the buttons that select the instrument being played.
        /// Each buttons has a different name, that will be used to get a different value of MIDI instrument using a Switch/Case.
        /// A Program Change message is created from that and the MIDI channel and the message is sent to the MIDI output device.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Selection_Click(object sender, RoutedEventArgs e)
        {
            // Define a byte to hold the MIDI instrument number to be used in the Program Change MIDI message
            byte midiInstrument;
            
            // Create a Button element to get the pressed instrument button
            Button instrumentNameButton = (Button)sender;

            // Create a string to hold the name of the pressed button to be used in a switch/case
            string instrumentName = instrumentNameButton.Name;

            // Depending on the name of the button pressed, set a new MIDI instrument and frequency interval for the instrument
            switch (instrumentName)
            {
                case "Piano":
                    // Set the MIDI instrument number to 0, Piano
                    midiInstrument = 0;

                    // Set the octaveInterval to 0 to set the frequency interval around the middle C
                    octaveInterval = 0;
                    break;
                case "Trombone":
                    // Set the MIDI instrument number to 57, Trombone (58-1)
                    midiInstrument = 57;

                    // Set the octaveInterval to -1 to set the frequency interval around one octave lower than the middle C
                    octaveInterval = -1;
                    break;
                case "Trumpet":
                    // Set the MIDI instrument number to 56, Trumpet (57-1)
                    midiInstrument = 56;

                    // Set the octaveInterval to 0 to set the frequency interval around the middle C
                    octaveInterval = 0;
                    break;
                case "Flute":
                    // Set the MIDI instrument number to 73, Trumpet (74-1)
                    midiInstrument = 73;

                    // Set the octaveInterval to 1 to set the frequency interval around one octave higher than the middle C
                    octaveInterval = 1;
                    break;

                // Default value will be equal to piano
                default:
                    // Set the MIDI instrument number to 0
                    midiInstrument = 0;

                    // Set the octaveInterval to 0 to set the frequency interval around the middle C
                    octaveInterval = 0;
                    break;
            }

            // Create the new Program Change message with the new selected instrument
            instrumentChange = new MidiProgramChangeMessage(channel, midiInstrument);

            // Send the Program Change midi message to port of the output midi device.
            midiOutPort.SendMessage(instrumentChange);
        }
    }
}
